#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import OrderedDict
import json

import dateutil
import pandas as pd
import pytest
import responses

from datarobot import Deployment
from datarobot.enums import ACCURACY_METRIC
from datarobot.models.deployment import Accuracy, AccuracyOverTime
from datarobot.utils import from_api


@pytest.fixture
def accuracy_response_data(deployment_data):
    return {
        "model_id": deployment_data["model"]["id"],
        "period": {"start": "2019-08-01T00:00:00.000000Z", "end": "2019-08-10T00:00:00.000000Z"},
        "metrics": {
            "AUC": {"baselineValue": 1, "value": 1.5, "percentChange": 50},
            "LogLoss": {"baselineValue": None, "value": None, "percentChange": None},
        },
    }


@pytest.fixture
def deployment_get_accuracy_response(unittest_endpoint, deployment_data, accuracy_response_data):
    url = "{}/deployments/{}/accuracy/".format(unittest_endpoint, deployment_data["id"])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_response_data),
    )


@pytest.fixture
def accuracy_over_time_response_data(deployment_data):
    return {
        "model_id": deployment_data["model"]["id"],
        "metric": "LogLoss",
        "summary": {
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-04T00:00:00.000000Z",
            },
            "value": 1,
            "sampleSize": 100,
        },
        "baseline": {"period": None, "value": None, "sampleSize": 0},
        "buckets": [
            {
                "period": {
                    "start": "2019-08-01T00:00:00.000000Z",
                    "end": "2019-08-02T00:00:00.000000Z",
                },
                "value": 0.22,
                "sampleSize": 30,
            },
            {
                "period": {
                    "start": "2019-08-02T00:00:00.000000Z",
                    "end": "2019-08-03T00:00:00.000000Z",
                },
                "value": 0.42,
                "sampleSize": 70,
            },
            {
                "period": {
                    "start": "2019-08-03T00:00:00.000000Z",
                    "end": "2019-08-04T00:00:00.000000Z",
                },
                "value": None,
                "sampleSize": 0,
            },
        ],
    }


@pytest.fixture
def deployment_get_accuracy_over_time_response(
    unittest_endpoint, deployment_data, accuracy_over_time_response_data
):
    url = "{}/deployments/{}/accuracyOverTime/".format(unittest_endpoint, deployment_data["id"])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_response_data),
    )


def assert_bucket(actual, expected):
    assert sorted(actual.keys()) == sorted(expected.keys())

    actual_period = actual["period"]
    if actual_period:
        assert actual["period"].keys() == expected["period"].keys()
        assert actual_period["start"] == dateutil.parser.parse(expected["period"]["start"])
        assert actual_period["end"] == dateutil.parser.parse(expected["period"]["end"])
    else:
        assert actual_period is None

    assert actual["value"] == expected["value"]
    assert actual["sample_size"] == expected["sample_size"]


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_get_accuracy_response")
def test_accuracy(deployment_data, accuracy_response_data):
    accuracy_objs = [
        Accuracy.get(deployment_data["id"]),
        Deployment.get(deployment_data["id"]).get_accuracy(),
    ]
    for accuracy in accuracy_objs:
        assert accuracy.model_id == accuracy_response_data["model_id"]

        expected_start_date = dateutil.parser.parse(accuracy_response_data["period"]["start"])
        expected_end_date = dateutil.parser.parse(accuracy_response_data["period"]["end"])
        assert accuracy.period["start"] == expected_start_date
        assert accuracy.period["end"] == expected_end_date

        assert accuracy.metrics.keys() == accuracy_response_data["metrics"].keys()
        for metric_name, metric in accuracy.metrics.items():
            expected_metric = from_api(
                accuracy_response_data["metrics"][metric_name], keep_null_keys=True
            )
            assert metric == expected_metric
            assert accuracy[metric_name] == expected_metric["value"]

        expected_metric_values = {
            name: metric["value"] for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.metric_values == expected_metric_values
        expected_metric_baseline = {
            name: metric["baselineValue"]
            for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.metric_baselines == expected_metric_baseline
        expected_percent_changes = {
            name: metric["percentChange"]
            for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.percent_changes == expected_percent_changes


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_get_accuracy_over_time_response")
def test_accuracy_over_time(deployment_data, accuracy_over_time_response_data):
    accuracy_objs = [
        AccuracyOverTime.get(deployment_data["id"]),
        Deployment.get(deployment_data["id"]).get_accuracy_over_time(),
    ]
    for accuracy_over_time in accuracy_objs:
        expected = from_api(accuracy_over_time_response_data, keep_null_keys=True)
        assert accuracy_over_time.metric == expected["metric"]
        assert accuracy_over_time.model_id == expected["model_id"]

        assert_bucket(accuracy_over_time.summary, expected["summary"])
        assert_bucket(accuracy_over_time.baseline, expected["baseline"])
        for index, bucket in enumerate(accuracy_over_time.buckets):
            assert_bucket(bucket, expected["buckets"][index])

        expected_bucket_values = [
            (dateutil.parser.parse(bucket["period"]["start"]), bucket["value"])
            for bucket in expected["buckets"]
        ]
        assert accuracy_over_time.bucket_values == OrderedDict(expected_bucket_values)

        expected_bucket_sample_sizes = [
            (dateutil.parser.parse(bucket["period"]["start"]), bucket["sample_size"])
            for bucket in expected["buckets"]
        ]
        assert accuracy_over_time.bucket_sample_sizes == OrderedDict(expected_bucket_sample_sizes)


@responses.activate
@pytest.mark.usefixtures("deployment_get_accuracy_over_time_response")
def test_accuracy_over_time_retrieve_as_dataframe(
    deployment_data, accuracy_over_time_response_data
):
    metric_names = [ACCURACY_METRIC.LOGLOSS, ACCURACY_METRIC.LOGLOSS]
    actual = AccuracyOverTime.get_as_dataframe(deployment_data["id"], metrics=metric_names)

    data = []
    start_times = []
    for bucket in accuracy_over_time_response_data["buckets"]:
        start_time = dateutil.parser.parse(bucket["period"]["start"])
        start_times.append(start_time)
        data.append([bucket["value"]] * 2)
    expected = pd.DataFrame(data, index=start_times, columns=metric_names)

    assert actual.equals(expected)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import Credential
from datarobot.utils import parse_time


@pytest.fixture
def credentials_server_resp():
    return {
        u"count": 2,
        u"data": [
            {
                u"creationDate": u"2020-02-03T10:13:27.411000Z",
                u"credentialId": u"5e37f247cf8a5f73842f6eab",
                u"credentialType": u"s3",
                u"description": u"",
                u"name": u"my_s3_cred_1",
            },
            {
                u"creationDate": u"2020-02-10T14:52:07.258000Z",
                u"credentialId": u"5e416e17cf8a5f6e7afcc52f",
                u"credentialType": u"basic",
                u"description": u"???",
                u"name": u"my_basic_cred_1",
            },
            {
                u"creationDate": u"2020-02-03T15:23:13.122000Z",
                u"credentialId": u"6e37f344cf8a5f73842f6eeb",
                u"credentialType": u"gcp",
                u"description": u"",
                u"name": u"my_gcp_cred_1",
            },
        ],
        u"next": None,
        u"previous": None,
        u"totalCount": 2,
    }


@pytest.fixture
def s3_credential_server_resp(credentials_server_resp):
    return credentials_server_resp["data"][0]


@pytest.fixture
def basic_credential_server_resp(credentials_server_resp):
    return credentials_server_resp["data"][1]


@pytest.fixture
def gcp_credential_server_resp(credentials_server_resp):
    return credentials_server_resp["data"][1]


@responses.activate
def test_list(credentials_endpoint, credentials_server_resp):
    responses.add(responses.GET, credentials_endpoint, json=credentials_server_resp)

    creds = Credential.list()

    for cred, server_payload in zip(creds, credentials_server_resp["data"]):
        assert cred.credential_id == server_payload["credentialId"]
        assert cred.name == server_payload["name"]
        assert cred.credential_type == server_payload["credentialType"]
        assert cred.creation_date == parse_time(server_payload["creationDate"])
        assert cred.description == server_payload["description"]


@responses.activate
def test_get(credentials_endpoint, s3_credential_server_resp):
    credential_id = s3_credential_server_resp["credentialId"]
    responses.add(
        responses.GET,
        "{}{}/".format(credentials_endpoint, credential_id),
        json=s3_credential_server_resp,
    )

    cred = Credential.get(credential_id)

    assert cred.credential_id == s3_credential_server_resp["credentialId"]
    assert cred.name == s3_credential_server_resp["name"]
    assert cred.credential_type == s3_credential_server_resp["credentialType"]
    assert cred.creation_date == parse_time(s3_credential_server_resp["creationDate"])
    assert cred.description == s3_credential_server_resp["description"]


@responses.activate
def test_create_basic(credentials_endpoint, basic_credential_server_resp):
    responses.add(responses.POST, credentials_endpoint, json=basic_credential_server_resp)

    cred = Credential.create_basic(
        name=basic_credential_server_resp["name"],
        user="USERNAME",
        password="PASSWORD",
        description=basic_credential_server_resp["description"],
    )

    assert cred.credential_id == basic_credential_server_resp["credentialId"]
    assert cred.name == basic_credential_server_resp["name"]
    assert cred.credential_type == basic_credential_server_resp["credentialType"]
    assert cred.creation_date == parse_time(basic_credential_server_resp["creationDate"])
    assert cred.description == basic_credential_server_resp["description"]


@responses.activate
def test_create_s3(credentials_endpoint, s3_credential_server_resp):
    responses.add(responses.POST, credentials_endpoint, json=s3_credential_server_resp)

    cred = Credential.create_s3(
        name=s3_credential_server_resp["name"],
        aws_access_key_id="?",
        aws_secret_access_key="!",
        description=s3_credential_server_resp["description"],
    )

    assert cred.credential_id == s3_credential_server_resp["credentialId"]
    assert cred.name == s3_credential_server_resp["name"]
    assert cred.credential_type == s3_credential_server_resp["credentialType"]
    assert cred.creation_date == parse_time(s3_credential_server_resp["creationDate"])
    assert cred.description == s3_credential_server_resp["description"]


@responses.activate
@pytest.mark.parametrize(
    "gcp_key",
    [
        pytest.param('{"token_uri": "foo", "client_email": "foo@example.org"}', id="key-as-str"),
        pytest.param({"token_uri": "foo", "client_email": "foo@example.org"}, id="key-as-dict"),
    ],
)
def test_create_gcp(credentials_endpoint, gcp_credential_server_resp, gcp_key):
    responses.add(responses.POST, credentials_endpoint, json=gcp_credential_server_resp)

    cred = Credential.create_gcp(
        name=gcp_credential_server_resp["name"],
        gcp_key=gcp_key,
        description=gcp_credential_server_resp["description"],
    )

    assert cred.credential_id == gcp_credential_server_resp["credentialId"]
    assert cred.name == gcp_credential_server_resp["name"]
    assert cred.credential_type == gcp_credential_server_resp["credentialType"]
    assert cred.creation_date == parse_time(gcp_credential_server_resp["creationDate"])
    assert cred.description == gcp_credential_server_resp["description"]

    assert json.loads(responses.calls[0].request.body)["gcpKey"] == {
        "token_uri": "foo",
        "client_email": "foo@example.org",
    }


def test_create_gcp_error():
    with pytest.raises(ValueError, match=r"^Could not parse gcp_key:.*$"):
        Credential.create_gcp(name="invalid credential", gcp_key="{invalid json", description="foo")


@responses.activate
def test_delete(credentials_endpoint, s3_credential_server_resp):
    credential_id = s3_credential_server_resp["credentialId"]
    responses.add(
        responses.GET,
        "{}{}/".format(credentials_endpoint, credential_id),
        json=s3_credential_server_resp,
    )
    responses.add(
        responses.DELETE, "{}{}/".format(credentials_endpoint, credential_id),
    )

    cred = Credential.get(credential_id)
    cred.delete()



# coding=utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import mock
import pytest
import responses

from datarobot._experimental import IdiomaticProject
from datarobot.enums import PROJECT_STAGE
from tests.utils import assert_raised_regex

# FIXTURES and HELPERS #############################################################################


@pytest.fixture
def feature_lists():
    return [
        {
            "id": "f-id-1",
            "projectId": "p-id",
            "name": "Raw Features",
            "features": ["One Fish", "Two Fish", "Red Fish", "Blue риба"],
        },
        {
            "id": "f-id-2",
            "projectId": "p-id",
            "name": "Informative Features",
            "features": ["One Fish", "Red Fish", "Blue риба"],
        },
        {
            "id": "f-id-3",
            "projectId": "p-id",
            "name": "Спеціальний список",
            "features": ["One Fish", "Blue риба"],
        },
    ]


@pytest.fixture
def new_feature_list():
    return {
        "id": "f-id-4",
        "projectId": "p-id",
        "name": "custom_list_1",
        "features": ["One Fish", "Blue риба"],
    }


@pytest.fixture
def idiomatic_project(project_with_target_data):
    return IdiomaticProject.from_data(project_with_target_data)


test_statuses = [
    {
        "autopilot_done": True,
        "stage": PROJECT_STAGE.MODELING,
        "stage_description": "Ready for modeling",
    },
    {"autopilot_done": False, "stage": PROJECT_STAGE.AIM, "stage_description": "Ready for aim"},
]


def test_time_series_not_supported(idiomatic_project):
    with mock.patch(
        "datarobot._experimental.IdiomaticProject.use_time_series",
        new_callable=mock.PropertyMock,
        return_value=True,
    ), pytest.raises(ValueError) as exc_info:
        idiomatic_project.fit()
        assert_raised_regex(exc_info, "Cannot currently run fit on timeseries projects")


@responses.activate
def test_fit_feature_lists(idiomatic_project, project_endpoint, feature_lists, new_feature_list):
    responses.add(
        responses.GET,
        "{}{}/featurelists/".format(project_endpoint, idiomatic_project.id),
        status=201,
        content_type="application/json",
        body=json.dumps(feature_lists),
    )

    responses.add(
        responses.POST,
        "{}{}/featurelists/".format(project_endpoint, idiomatic_project.id),
        status=201,
        content_type="application/json",
        body=json.dumps(new_feature_list),
    )

    with mock.patch.object(idiomatic_project, "start_autopilot"), mock.patch.object(
        idiomatic_project, "refresh"
    ), mock.patch.object(
        idiomatic_project, "get_status", return_value={"stage": PROJECT_STAGE.AIM}
    ), mock.patch.object(
        idiomatic_project, "set_target"
    ) as mock_set_target:
        # call by feature list id
        idiomatic_project.fit(feature_list="f-id-1")
        assert mock_set_target.called is True
        assert mock_set_target.call_args.kwargs["featurelist_id"] == "f-id-1"
        # call by feature list name -- default
        mock_set_target.reset_mock()
        idiomatic_project.fit(feature_list="Informative Features")
        assert mock_set_target.called is True
        assert mock_set_target.call_args.kwargs["featurelist_id"] == "f-id-2"
        # call by feature list name -- custom
        mock_set_target.reset_mock()
        idiomatic_project.fit(feature_list="Спеціальний список")
        assert mock_set_target.called is True
        assert mock_set_target.call_args.kwargs["featurelist_id"] == "f-id-3"
        # call by list of features
        mock_set_target.reset_mock()
        idiomatic_project.fit(feature_list=["One Fish", "Blue риба"])
        assert mock_set_target.called is True
        assert mock_set_target.call_args.kwargs["featurelist_id"] == "f-id-4"


@responses.activate
@pytest.mark.parametrize("status", test_statuses)
def test_fit(idiomatic_project, project_endpoint, feature_lists, status):
    responses.add(
        responses.GET,
        "{}{}/featurelists/".format(project_endpoint, idiomatic_project.id),
        status=201,
        content_type="application/json",
        body=json.dumps(feature_lists),
    )

    responses.add(
        responses.GET,
        "{}{}/status/".format(project_endpoint, idiomatic_project.id),
        status=201,
        content_type="application/json",
        body=json.dumps(status),
    )

    with mock.patch.object(idiomatic_project, "set_target") as mock_set_target, mock.patch.object(
        idiomatic_project, "start_autopilot"
    ) as mock_start_autopilot, mock.patch.object(idiomatic_project, "refresh"):
        idiomatic_project.fit()
        assert mock_set_target.called is not status["autopilot_done"]
        assert mock_start_autopilot.called is status["autopilot_done"]



# -*- coding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import numpy as np
import pandas as pd
import pytest
import responses
import six
from trafaret import DataError

from datarobot._experimental.models.model import CombinedModel
from datarobot._experimental.models.project import Project
from tests.utils import add_response

_project_id = "projectId"
_target = "target"
_model_id = "modelId"
_base_url = "https://host_name.com/projects/{}/combinedModels/".format(_project_id)
_valid_combined_model_id = "validCombinedModelId"
_invalid_combined_model_id = "validCombinedModelId"
_valid_combined_model_url = "{0}{1}/".format(_base_url, _valid_combined_model_id)
_invalid_combined_model_url = "{0}{1}/".format(_base_url, _invalid_combined_model_id)
_combined_model_segments_dl_url = "{0}{1}/segments/download/".format(
    _base_url, _valid_combined_model_id
)
_segment_champion_url = "https://host_name.com/projects/{}/segmentChampion/".format(_project_id)


@pytest.fixture
def combined_model_record():
    return {
        "combinedModelId": _valid_combined_model_id,
        "projectId": _project_id,
        "segmentationTaskId": "segmentationTaskId",
    }


@pytest.fixture
def combined_model_list_response(combined_model_record):
    return {"count": 1, "next": None, "previous": None, "data": [combined_model_record]}


@pytest.fixture
def combined_model_get_json(combined_model_record):
    return json.dumps(combined_model_record)


@pytest.fixture
def invalid_json():
    return json.dumps({"message": "Not Found"})


@pytest.fixture
def combined_model_list_json(combined_model_list_response):
    return json.dumps(combined_model_list_response)


@pytest.fixture
def mock_segment_csv():
    row_indexes = [u"row_{}".format(i) for i in range(10)]
    column_names = [u"カテゴリー_{}".format(i) for i in range(5)]
    rs = np.random.RandomState(1234)
    df = pd.DataFrame(
        data=rs.uniform(-1, 1, (len(row_indexes), len(column_names))),
        columns=column_names,
        index=row_indexes,
    )
    return df.to_csv(encoding="utf-8")


@responses.activate
def test_combined_model_list(combined_model_list_json):
    responses.add(
        responses.GET,
        _base_url,
        body=combined_model_list_json,
        status=200,
        content_type="application/json",
    )
    project = Project(id=_project_id)
    combined_models = project.get_combined_models()
    assert len(combined_models) == 1
    combined_model = combined_models[0]

    assert isinstance(combined_model, CombinedModel)
    assert combined_model.id == _valid_combined_model_id
    assert combined_model.project_id == _project_id


@responses.activate
def test_combined_model_get_valid(combined_model_get_json):
    responses.add(
        responses.GET,
        _valid_combined_model_url,
        body=combined_model_get_json,
        status=200,
        content_type="application/json",
    )
    combined_model = CombinedModel.get(_project_id, _valid_combined_model_id)

    assert isinstance(combined_model, CombinedModel)
    assert combined_model.id == _valid_combined_model_id
    assert combined_model.project_id == _project_id


@responses.activate
def test_combined_model_invalid_get(invalid_json):
    responses.add(responses.GET, _invalid_combined_model_url, body=invalid_json)
    with pytest.raises(DataError):
        CombinedModel.get(_project_id, _invalid_combined_model_id)


@responses.activate
def test_combined_model_get_segments_as_dataframe(mock_segment_csv):
    add_response(_combined_model_segments_dl_url, mock_segment_csv, content_type="text/csv")
    combined_model = CombinedModel(_valid_combined_model_id, _project_id)
    df = combined_model.get_segments_as_dataframe()

    assert isinstance(df, pd.DataFrame)
    assert df.shape == (10, 5)


@responses.activate
def test_combined_model_get_segments_as_csv(mock_segment_csv):
    add_response(_combined_model_segments_dl_url, mock_segment_csv, content_type="text/csv")
    combined_model = CombinedModel(_valid_combined_model_id, _project_id)
    file_buffer = six.StringIO()
    combined_model.get_segments_as_csv(file_buffer)
    file_buffer.seek(0)

    # Assert the values contained in the output reasonably match the inputs
    actual_values = pd.read_csv(six.StringIO(file_buffer.read())).values
    expected_values = pd.read_csv(six.StringIO(mock_segment_csv)).values[:, 1:]
    np.testing.assert_almost_equal(actual_values, expected_values)


@responses.activate
def test_combined_model_valid_set_segment_champion(invalid_json):
    responses.add(
        responses.PUT,
        _segment_champion_url,
        body=json.dumps({"combinedModelId": str(_model_id)}),
        status=200,
        content_type="application/json",
    )

    combined_model = CombinedModel.set_segment_champion(_project_id, _model_id)
    assert combined_model == "modelId"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest

from datarobot import Client
from datarobot.client import get_client


@pytest.mark.usefixtures("no_client_version_check")
def test_token_endpoint_overrides_config_path(config_path):
    Client(token="code_token", endpoint="https://code_endpoint.com", config_path=config_path)
    client = get_client()
    assert client.token == "code_token"
    assert client.endpoint == "https://code_endpoint.com"


@pytest.mark.usefixtures("env_config_path", "no_client_version_check")
def test_config_overrides_env_variables(config_path, code_config_token, code_config_endpoint):
    Client(config_path=config_path)
    client = get_client()
    assert client.token == code_config_token
    assert client.endpoint == code_config_endpoint


@pytest.mark.usefixtures("env_config_path", "env_config_vars", "no_client_version_check")
def test_env_config_overrides_env_vars(env_config_endpoint, env_config_token):
    client = get_client()
    assert client.token == env_config_token
    assert client.endpoint == env_config_endpoint


@pytest.mark.usefixtures("env_config_vars", "default_config", "no_client_version_check")
def test_config_vars_override_config_file_in_default_location(env_token, env_endpoint):
    client = get_client()
    assert client.token == env_token
    assert client.endpoint == env_endpoint


@pytest.mark.usefixtures("default_config", "no_client_version_check")
def test_default_config_if_nothing_else(default_config_endpoint, default_config_token):
    client = get_client()
    assert client.token == default_config_token
    assert client.endpoint == default_config_endpoint


@pytest.mark.usefixtures("no_client_version_check")
def test_client_ssl_verify_default_true():
    client = Client(token="token", endpoint="https://code_endpoint.com")
    assert client.verify is True


@pytest.mark.usefixtures("no_client_version_check")
def test_client_ssl_verify_false():
    client = Client(token="token", endpoint="https://code_endpoint.com", ssl_verify=False)
    assert client.verify is False


@pytest.mark.usefixtures("no_client_version_check")
def test_client_use_default_token_type():
    client = Client(token="token", endpoint="https://code_endpoint.com")
    assert client.token_header == {"Authorization": "Token token"}


@pytest.mark.usefixtures("no_client_version_check")
def test_client_use_bearer_token_type():
    client = Client(token="token", token_type="Bearer", endpoint="https://code_endpoint.com")
    assert client.token_header == {"Authorization": "Bearer token"}



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from copy import deepcopy
import json

import pytest
import responses

from datarobot import ExecutionEnvironment
from datarobot.models.execution_environment import RequiredMetadataKey
from datarobot.utils import camelize, underscorize
from tests.model.utils import assert_version


@pytest.fixture
def mocked_environments(mocked_execution_environments):
    return mocked_execution_environments


@pytest.fixture
def mocked_environment(mocked_execution_environment):
    return mocked_execution_environment


@pytest.fixture
def make_environments_url(unittest_endpoint):
    def _make_environments_url(environment_id=None):
        base_url = "{}/executionEnvironments/".format(unittest_endpoint)
        if environment_id is not None:
            return "{}{}/".format(base_url, environment_id)
        return base_url

    return _make_environments_url


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def assert_environment(environment, environment_json):
    assert environment.id == environment_json["id"]
    assert environment.name == environment_json["name"]
    assert environment.description == environment_json["description"]
    assert environment.programming_language == environment_json["programmingLanguage"]
    assert environment.is_public == environment_json["isPublic"]
    assert environment.created_at == environment_json["created"]
    if environment.latest_version is None:
        assert environment_json["latestVersion"] is None
    else:
        assert_version(environment.latest_version, environment_json["latestVersion"])

    if environment_json.get("requiredMetadataKeys"):
        assert all(
            key.field_name == key_json["fieldName"] and key.display_name == key_json["displayName"]
            for key, key_json in zip(
                environment.required_metadata_keys, environment_json["requiredMetadataKeys"]
            )
        )


def test_from_server_data(mocked_environment):
    environment = ExecutionEnvironment.from_server_data(mocked_environment)
    assert_environment(environment, mocked_environment)


@responses.activate
@pytest.mark.parametrize(
    "mock_environment",
    ["mocked_environment", "mocked_execution_environment_with_required_metadata_keys"],
)
def test_create_environment(request, make_environments_url, tmpdir, mock_environment):
    mock_environment = request.getfixturevalue(mock_environment)
    responses.add(
        responses.POST,
        make_environments_url(),
        status=200,
        content_type="application/json",
        body=json.dumps({"id": mock_environment["id"]}),
    )
    responses.add(
        responses.GET,
        make_environments_url(mock_environment["id"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mock_environment),
    )

    keys = [
        RequiredMetadataKey.from_data({underscorize(k): v for k, v in key.items()})
        for key in mock_environment.get("requiredMetadataKeys", {})
    ]
    keys = keys or None

    environment = ExecutionEnvironment.create(
        mock_environment["name"], mock_environment["description"], required_metadata_keys=keys
    )
    assert_environment(environment, mock_environment)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_environments_url()
    if mock_environment.get("requiredMetadataKeys"):
        assert json.loads(responses.calls[0].request.body)[
            "requiredMetadataKeys"
        ] == mock_environment.get("requiredMetadataKeys")
    else:
        assert "requiredMetadataKeys" not in json.loads(responses.calls[0].request.body)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_environments_url(mock_environment["id"])


@responses.activate
@pytest.mark.parametrize(
    "mock_environment",
    ["mocked_environment", "mocked_execution_environment_with_required_metadata_keys"],
)
def test_get_environment(request, mock_environment, make_environments_url):
    mock_environment = request.getfixturevalue(mock_environment)
    url = make_environments_url(mock_environment["id"])
    mock_get_response(url, mock_environment)

    environment = ExecutionEnvironment.get(mock_environment["id"])
    assert_environment(environment, mock_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_environments(mocked_environments, make_environments_url):
    url = make_environments_url()
    mock_get_response(url, mocked_environments)

    environments = ExecutionEnvironment.list()

    assert len(environments) == len(mocked_environments["data"])
    for environment, mocked_environment in zip(environments, mocked_environments["data"]):
        assert_environment(environment, mocked_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_environments_multiple_pages(mocked_environments, make_environments_url):
    url1 = make_environments_url()
    url2 = make_environments_url() + "2"

    mocked_environments_2nd = deepcopy(mocked_environments)
    mocked_environments["next"] = url2

    mock_get_response(url1, mocked_environments)
    mock_get_response(url2, mocked_environments_2nd)

    environments = ExecutionEnvironment.list()
    assert len(environments) == len(mocked_environments["data"]) + len(
        mocked_environments_2nd["data"]
    )
    for environment, mocked_environment in zip(
        environments, mocked_environments["data"] + mocked_environments_2nd["data"]
    ):
        assert_environment(environment, mocked_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_list_filter_search_for(mocked_environment, make_environments_url):
    url = make_environments_url() + "?searchFor=xyz"
    mock_get_response(
        url, {"count": 1, "next": None, "previous": None, "data": [mocked_environment]}
    )

    environments = ExecutionEnvironment.list(search_for="xyz")

    assert len(environments) == 1
    assert_environment(environments[0], mocked_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_delete(mocked_environment, make_environments_url):
    url = make_environments_url(mocked_environment["id"])

    mock_get_response(url, mocked_environment)
    responses.add(
        responses.DELETE, url, status=204, content_type="application/json",
    )

    execution_environment = ExecutionEnvironment.get(mocked_environment["id"])
    execution_environment.delete()

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "DELETE"
    assert responses.calls[1].request.url == url


@responses.activate
@pytest.mark.parametrize(
    "attrs",
    [
        {"name": "nx"},
        {"description": "dx"},
        {"name": "nx", "description": "dx"},
        {
            "requiredMetadataKeys": [
                RequiredMetadataKey(
                    field_name="FIELD_{}".format(i), display_name="field {}".format(i)
                )
                for i in range(3)
            ]
        },
    ],
)
def test_update(mocked_environment, make_environments_url, attrs):
    envrironment_id = mocked_environment["id"]
    url = make_environments_url(envrironment_id)

    mock_get_response(url, mocked_environment)

    mock_update_fields = attrs.copy()
    if "requiredMetadataKeys" in mock_update_fields:
        mock_update_fields["requiredMetadataKeys"] = [
            {camelize(k): v for k, v in key.to_dict().items()}
            for key in mock_update_fields["requiredMetadataKeys"]
        ]

    mocked_environment.update(**mock_update_fields)

    responses.add(
        responses.PATCH,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_environment),
    )

    environment = ExecutionEnvironment.get(envrironment_id)
    environment.update(**{underscorize(k): v for k, v in attrs.items()})
    assert_environment(environment, mocked_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "PATCH"
    assert responses.calls[1].request.url == url
    assert json.loads(responses.calls[1].request.body) == mock_update_fields


@responses.activate
def test_refresh(mocked_environment, make_environments_url):
    envrironment_id = mocked_environment["id"]
    url = make_environments_url(envrironment_id)

    mock_get_response(url, mocked_environment)

    mocked_environment.update({"programmingLanguage": "r", "name": "xx"})

    mock_get_response(url, mocked_environment)

    environment = ExecutionEnvironment.get(envrironment_id)
    environment.refresh()
    assert_environment(environment, mocked_environment)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import mock
import responses
import six

from datarobot import Connector


@responses.activate
def test_list(connectors_endpoint, connectors_list_server_resp):
    responses.add(responses.GET, connectors_endpoint, json=connectors_list_server_resp)
    connectors = Connector.list()
    for connector, server_payload in zip(connectors, connectors_list_server_resp["data"]):
        assert connector.id == server_payload["id"]
        assert connector.creator == server_payload["creatorId"]
        assert connector.configuration_id == server_payload["configurationId"]
        assert connector.base_name == server_payload["baseName"]
        assert connector.canonical_name == server_payload["canonicalName"]


@responses.activate
def test_get(connectors_endpoint, s3_connector_server_resp):
    connector_id = s3_connector_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
    )

    connector = Connector.get(connector_id)
    assert connector.id == s3_connector_server_resp["id"]
    assert connector.creator == s3_connector_server_resp["creatorId"]
    assert connector.configuration_id == s3_connector_server_resp["configurationId"]
    assert connector.base_name == s3_connector_server_resp["baseName"]
    assert connector.canonical_name == s3_connector_server_resp["canonicalName"]


@responses.activate
def test_future_proof(connectors_endpoint, s3_connector_server_resp):
    connector_id = s3_connector_server_resp["id"]
    resp = dict(s3_connector_server_resp, new_key="new")
    responses.add(responses.GET, "{}{}/".format(connectors_endpoint, connector_id), json=resp)

    connector = Connector.get(connector_id)
    assert connector.id == connector_id


@responses.activate
def test_create(
    connectors_endpoint,
    driver_libraries_endpoint,
    driver_library_upload_server_resp,
    s3_connector_server_resp,
):
    connector_id = s3_connector_server_resp["id"]
    responses.add(responses.POST, driver_libraries_endpoint, json=driver_library_upload_server_resp)
    responses.add(
        responses.POST,
        connectors_endpoint,
        json=s3_connector_server_resp,
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    location = "https://host_name.com/externalConnectors/{}/".format(connector_id)
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": location},
    )
    responses.add(
        responses.GET,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
    )

    with mock.patch("os.path.exists"):
        with mock.patch("datarobot.rest.open", create=True) as open_mock:
            open_mock.return_value = six.StringIO("thra\ntata\nrata")
            connector = Connector.create("/tmp/{}".format(s3_connector_server_resp["baseName"]))

    assert connector.id == s3_connector_server_resp["id"]
    assert connector.creator == s3_connector_server_resp["creatorId"]
    assert connector.configuration_id == s3_connector_server_resp["configurationId"]
    assert connector.base_name == s3_connector_server_resp["baseName"]
    assert connector.canonical_name == s3_connector_server_resp["canonicalName"]


@responses.activate
def test_update(
    connectors_endpoint,
    driver_libraries_endpoint,
    driver_library_upload_server_resp,
    s3_connector_server_resp,
):
    connector_id = s3_connector_server_resp["id"]

    responses.add(
        responses.GET,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
    )
    responses.add(responses.POST, driver_libraries_endpoint, json=driver_library_upload_server_resp)
    responses.add(
        responses.PATCH,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    location = "https://host_name.com/externalConnectors/{}/".format(connector_id)
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": location},
    )
    responses.add(
        responses.GET,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
    )

    with mock.patch("os.path.exists"):
        with mock.patch("datarobot.rest.open", create=True) as open_mock:
            open_mock.return_value = six.StringIO("thra\ntata\nrata")
            connector = Connector.get(connector_id)
            connector = connector.update("/tmp/{}".format(s3_connector_server_resp["baseName"]))

    assert connector.id == s3_connector_server_resp["id"]
    assert connector.creator == s3_connector_server_resp["creatorId"]
    assert connector.configuration_id == s3_connector_server_resp["configurationId"]
    assert connector.base_name == s3_connector_server_resp["baseName"]
    assert connector.canonical_name == s3_connector_server_resp["canonicalName"]


@responses.activate
def test_delete(connectors_endpoint, s3_connector_server_resp):
    connector_id = s3_connector_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(connectors_endpoint, connector_id),
        json=s3_connector_server_resp,
    )
    delete_url = "{}{}/".format(connectors_endpoint, connector_id)
    responses.add(responses.DELETE, delete_url)

    driver = Connector.get(connector_id)
    driver.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import responses

from datarobot import FrozenModel, Project


@responses.activate
def test_get_frozen_models_returns_valid_objects(frozen_models_list_response):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/frozenModels/",
        json=frozen_models_list_response,
        status=200,
    )
    frozen_models = list(Project("p-id").get_frozen_models())
    assert len(frozen_models) == 2
    assert all(isinstance(frozen, FrozenModel) for frozen in frozen_models)



# -*- coding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json
from operator import itemgetter

import pytest
import responses
from trafaret import DataError

from datarobot import errors
from datarobot.models import FeatureAssociationMatrix
from datarobot.utils import from_api


@responses.activate
def test_get(project_id, feature_association_matrix_url, feature_association_matrix):
    responses.add(
        responses.GET,
        feature_association_matrix_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix),
    )
    feature_association = FeatureAssociationMatrix.get(project_id=project_id)
    assert feature_association
    assert feature_association.project_id == project_id
    assert len(feature_association.strengths) == len(feature_association_matrix["strengths"])
    assert len(feature_association.features) == len(feature_association_matrix["features"])
    assert sorted(feature_association.strengths, key=itemgetter("feature1")) == sorted(
        feature_association_matrix["strengths"], key=itemgetter("feature1")
    )
    assert sorted(feature_association.features, key=itemgetter("feature")) == sorted(
        from_api(feature_association_matrix["features"]), key=itemgetter("feature")
    )


@responses.activate
def test_get_for_featurelist(
    project_id, feature_association_matrix_url, feature_association_matrix
):
    featurelist_id = "6033e2b9ecb9108111e4c677"
    url = feature_association_matrix_url + "?featurelistId={}".format(featurelist_id)
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix),
    )
    feature_association = FeatureAssociationMatrix.get(
        project_id=project_id, featurelist_id=featurelist_id
    )
    assert feature_association
    assert feature_association.project_id == project_id
    assert len(feature_association.strengths) == len(feature_association_matrix["strengths"])
    assert len(feature_association.features) == len(feature_association_matrix["features"])
    assert sorted(feature_association.strengths, key=itemgetter("feature1")) == sorted(
        feature_association_matrix["strengths"], key=itemgetter("feature1")
    )
    assert sorted(feature_association.features, key=itemgetter("feature")) == sorted(
        from_api(feature_association_matrix["features"]), key=itemgetter("feature")
    )


@responses.activate
def test_get__not_found(project_id, feature_association_matrix_url):
    responses.add(responses.GET, feature_association_matrix_url, status=404)
    with pytest.raises(errors.ClientError, match="404 client error: "):
        FeatureAssociationMatrix.get(project_id=project_id)


@pytest.mark.parametrize("project_id", [None, "", 1, 42.1])
def test_get__invalid_project_id(project_id):
    with pytest.raises(DataError):
        FeatureAssociationMatrix.get(project_id)


def test_get__invalid_arguments(project_id):
    with pytest.raises(DataError):
        FeatureAssociationMatrix.get(
            project_id, metric="fake", association_type="fake", featurelist_id="",
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import defaultdict
from copy import deepcopy
import json
import os
import uuid

import mock
import pytest
from requests_toolbelt import MultipartEncoder
import responses
import six

from datarobot import CustomModelVersion
from datarobot.models.custom_model_version import RequiredMetadataValue
from datarobot.utils import camelize
from tests.model.utils import assert_custom_model_version, fields_to_dict


@pytest.fixture
def mocked_version(mocked_custom_model_version):
    return mocked_custom_model_version


@pytest.fixture
def mocked_version_with_resources(mocked_custom_model_version_with_resources):
    return mocked_custom_model_version_with_resources


@pytest.fixture
def make_versions_url(unittest_endpoint):
    def _make_versions_url(environment_id, version_id=None):
        base_url = "{}/customModels/{}/versions/".format(unittest_endpoint, environment_id)
        if version_id is not None:
            return "{}{}/".format(base_url, version_id)
        return base_url

    return _make_versions_url


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def test_from_server_data(mocked_version):
    version = CustomModelVersion.from_server_data(mocked_version)
    assert_custom_model_version(version, mocked_version)


@responses.activate
@pytest.mark.parametrize(
    "version_json",
    [
        "mocked_version",
        "mocked_custom_model_version_no_dependencies",
        "mocked_custom_model_version_no_base_environment",
        "mocked_custom_model_version_future_field_in_version",
        "mocked_custom_model_version_future_field_in_dependency",
        "mocked_custom_model_version_future_field_in_constraint",
        "mocked_custom_model_version_with_required_metadata",
    ],
)
def test_get_version(request, version_json, make_versions_url):
    # arrange
    version_json = request.getfixturevalue(version_json)
    url = make_versions_url(version_json["customModelId"], version_json["id"])
    mock_get_response(url, version_json)

    # act
    version = CustomModelVersion.get(version_json["customModelId"], version_json["id"])

    # assert
    assert_custom_model_version(version, version_json)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions(mocked_versions, make_versions_url):
    # arrange
    custom_model_id = mocked_versions["data"][0]["customModelId"]
    url = make_versions_url(custom_model_id)
    mock_get_response(url, mocked_versions)

    # act
    versions = CustomModelVersion.list(custom_model_id)

    # assert
    assert len(versions) == len(mocked_versions["data"])
    for version, mocked_version in zip(versions, mocked_versions["data"]):
        assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions_multiple_pages(mocked_versions, make_versions_url):
    # arrange
    custom_model_id = mocked_versions["data"][0]["customModelId"]

    url1 = make_versions_url(custom_model_id)
    url2 = make_versions_url(custom_model_id) + "2"

    mocked_versions_2nd = deepcopy(mocked_versions)
    mocked_versions["next"] = url2

    mock_get_response(url1, mocked_versions)
    mock_get_response(url2, mocked_versions_2nd)

    # act
    versions = CustomModelVersion.list(custom_model_id)

    # assert
    assert len(versions) == len(mocked_versions["data"]) + len(mocked_versions_2nd["data"])
    for version, mocked_version in zip(
        versions, mocked_versions["data"] + mocked_versions_2nd["data"]
    ):
        assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url1
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url2


@responses.activate
def test_download(mocked_version, make_versions_url, tmpdir):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)
    responses.add(
        responses.GET,
        url + "download/",
        status=200,
        content_type="application/json",
        body=b"content",
    )

    downloaded_file = tmpdir.mkdir("sub").join("download")
    downloaded_file_path = str(downloaded_file)

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    version.download(downloaded_file_path)

    # assert
    assert six.ensure_binary(downloaded_file.read()) == b"content"

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "download/"


@responses.activate
@pytest.mark.parametrize("use_deprecated_metadata", [True, False])
def test_update(
    mocked_custom_model_version_with_required_metadata, make_versions_url, use_deprecated_metadata,
):
    mocked_version = mocked_custom_model_version_with_required_metadata
    # arrange
    attrs = {"description": "xx"}
    if use_deprecated_metadata:
        attrs.update({"required_metadata": {"REQUIRED_FIELD": "super important value"}})
    else:
        attrs.update(
            {
                "required_metadata_values": [
                    RequiredMetadataValue(
                        field_name="REQUIRED_FIELD", value="super important value"
                    )
                ]
            }
        )
    api_attrs = {camelize(k): v for k, v in attrs.items()}

    if not use_deprecated_metadata:
        api_attrs["requiredMetadataValues"] = [
            {camelize(k): v for k, v in val.to_dict().items()}
            for val in attrs["required_metadata_values"]
        ]

    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    mocked_version.update(api_attrs)

    responses.add(
        responses.PATCH,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    version.update(**attrs)

    # assert
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "PATCH"
    assert responses.calls[1].request.url == url
    assert responses.calls[1].request.body == json.dumps(api_attrs).encode()


@responses.activate
@pytest.mark.parametrize("required_metadata", [{}, {"a": "b"}], ids=["empty", "non-empty"])
@pytest.mark.parametrize(
    "required_metadata_values",
    [[], [RequiredMetadataValue(field_name="a", value="b")]],
    ids=["empty", "non-empty"],
)
def test_update_with_both_required_metadata_and_required_metadata_value_raises_error(
    mocked_custom_model_version_with_required_metadata,
    make_versions_url,
    required_metadata,
    required_metadata_values,
):
    mocked_version = mocked_custom_model_version_with_required_metadata
    # arrange
    attrs = {
        "description": "xx",
        "required_metadata": required_metadata,
        "required_metadata_values": required_metadata_values,
    }

    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    with pytest.raises(ValueError):
        version.update(**attrs)


@responses.activate
def test_refresh(mocked_version, make_versions_url):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    mocked_version.update({"description": "xx"})

    mock_get_response(url, mocked_version)

    # act
    version = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    version.refresh()

    # assert
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url


@responses.activate
@pytest.mark.parametrize(
    "mocked_fixture_name",
    [
        "mocked_version",
        "mocked_version_with_resources",
        "mocked_custom_model_version_with_required_metadata",
    ],
)
@pytest.mark.parametrize("use_deprecated_metadata", [True, False])
def test_create_clean(
    request,
    mocked_fixture_name,
    make_versions_url,
    tmpdir,
    base_environment_id,
    use_deprecated_metadata,
):
    mocked_version = request.getfixturevalue(mocked_fixture_name)

    url = make_versions_url(mocked_version["customModelId"])

    responses.add(
        responses.POST,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    file = tmpdir.mkdir("sub").join("file.txt")
    file.write(b"content")
    file_path = str(file)

    attrs = dict(
        custom_model_id=mocked_version["customModelId"],
        base_environment_id=base_environment_id,
        is_major_update=True,
        files=[(file_path, "/d/file.txt")],
        network_egress_policy=mocked_version["networkEgressPolicy"],
        maximum_memory=mocked_version["maximumMemory"],
        replicas=mocked_version["replicas"],
    )
    api_attrs = {
        "baseEnvironmentId": [base_environment_id],
        "isMajorUpdate": [str(True)],
        "file": [(os.path.basename(file_path), mock.ANY)],
        "filePath": ["/d/file.txt"],
    }

    if mocked_version["networkEgressPolicy"]:
        api_attrs["networkEgressPolicy"] = [mocked_version["networkEgressPolicy"]]
    if mocked_version["maximumMemory"]:
        api_attrs["maximumMemory"] = [str(mocked_version["maximumMemory"])]
    if mocked_version["replicas"]:
        api_attrs["replicas"] = [str(mocked_version["replicas"])]

    if mocked_version.get("requiredMetadata"):
        if use_deprecated_metadata:
            attrs.update({"required_metadata": mocked_version.get("requiredMetadata")})
            api_attrs.update(
                {"requiredMetadata": [json.dumps(mocked_version.get("requiredMetadata"))]}
            )
        else:
            attrs.update(
                {
                    "required_metadata_values": [
                        RequiredMetadataValue.from_server_data(val)
                        for val in mocked_version.get("requiredMetadataValues")
                    ]
                }
            )
            api_attrs.update(
                {
                    "requiredMetadataValues": [
                        json.dumps(mocked_version.get("requiredMetadataValues"))
                    ],
                }
            )

    version = CustomModelVersion.create_clean(**attrs)
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url
    request_body = responses.calls[0].request.body
    field_dict = fields_to_dict(request_body)
    field_dict["file"] = [(path, mock.ANY) for path, _ in field_dict["file"]]
    assert field_dict == api_attrs


@pytest.mark.parametrize("required_metadata", [{}, {"a": "b"}], ids=["empty", "non-empty"])
@pytest.mark.parametrize(
    "required_metadata_values",
    [[], [RequiredMetadataValue(field_name="a", value="b")]],
    ids=["empty", "non-empty"],
)
def test_create_clean_with_both_required_metadata_and_required_metadata_value_raises_error(
    required_metadata, required_metadata_values
):
    attrs = dict(
        custom_model_id="abc123",
        base_environment_id="dev456",
        is_major_update=True,
        required_metadata=required_metadata,
        required_metadata_values=required_metadata_values,
    )

    with pytest.raises(ValueError):
        CustomModelVersion.create_clean(**attrs)


@responses.activate
def test_create_from_previous(mocked_version, make_versions_url, tmpdir, base_environment_id):
    url = make_versions_url(mocked_version["customModelId"])

    responses.add(
        responses.PATCH,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )

    file = tmpdir.mkdir("sub").join("file")
    file.write(b"content")
    file_path = str(file)

    version = CustomModelVersion.create_from_previous(
        mocked_version["customModelId"],
        base_environment_id,
        True,
        files=[(file_path, "/d/file.txt")],
        files_to_delete=["2323423423423423"],
    )
    assert_custom_model_version(version, mocked_version)

    assert responses.calls[0].request.method == "PATCH"
    assert responses.calls[0].request.url == url


@responses.activate
@pytest.mark.parametrize(
    "method",
    [
        (CustomModelVersion.create_clean, "POST"),
        (CustomModelVersion.create_from_previous, "PATCH"),
    ],
)
@pytest.mark.parametrize(
    "expected_folder_paths",
    [None, [], ["a.zip", "a/a.zip", "sub/a.txt", "sub/b.pdf", "sub/sub2/c.rtf"]],
)
@pytest.mark.parametrize("expected_file_paths", [None, [], ["y.xyz", "dd/ttt.txt"]])
def test_create_folder_and_files(
    mocked_version,
    make_versions_url,
    tmpdir,
    method,
    expected_folder_paths,
    expected_file_paths,
    base_environment_id,
):
    custom_model_version_method = method[0]
    http_method = method[1]

    # arrange
    responses.add(
        http_method,
        make_versions_url(mocked_version["customModelId"]),
        status=200,
        body=json.dumps(mocked_version),
    )

    # arrange: create folder with required structure to be uploaded
    folder_path = None
    if expected_folder_paths:
        folder_path = tmpdir.mkdir(str(uuid.uuid4()))
        for expected_path in expected_folder_paths:
            path_items, file = os.path.split(expected_path)
            d = folder_path
            if path_items:
                for path_item in path_items.split("/"):
                    if os.path.exists(os.path.join(str(d), path_item)):
                        d = d.join(path_item)
                    else:
                        d = d.mkdir(path_item)
            d.join(file).write(b"content")
        folder_path = str(folder_path)

    # arrange: create files to be uploaded
    files = None
    if expected_file_paths:
        files = []
        for expected_path in expected_file_paths:
            file = tmpdir.join(str(uuid.uuid4()))
            file.write(b"content")
            files.append((str(file), expected_path))

    # act
    custom_model_version_method(
        mocked_version["customModelId"],
        base_environment_id,
        True,
        folder_path=folder_path,
        files=files,
    )

    # assert
    req = responses.calls[0].request
    assert req.method == http_method
    assert req.url == make_versions_url(mocked_version["customModelId"])
    assert isinstance(req.body, MultipartEncoder)

    # get fields submitted with request
    fields = defaultdict(list)
    for name, value in req.body.fields:
        fields[name].append(value)

    assert len(fields["isMajorUpdate"]) == 1
    assert fields["isMajorUpdate"][0] == "True"

    files = fields["file"]
    file_paths = fields["filePath"]

    # verify that files & filePaths make pairs
    assert len(files) == len(file_paths)
    for file, file_path in zip(files, file_paths):
        assert file[0] == os.path.basename(file_path)

    expected_folder_paths = expected_folder_paths or []
    expected_file_paths = expected_file_paths or []

    # There's an option to create a custom model version by passing a pass to a directory.
    # For such a directory, files are created with expected_folder_paths.
    # As there's no control in what order filesystem reads files from a such a directory,
    # convert paths to set and then assert
    assert set(file_paths) == set(expected_folder_paths) | set(expected_file_paths)


@responses.activate
def test_get_feature_impact(mocked_version, make_versions_url, feature_impact_server_data):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)
    responses.add(
        responses.GET,
        url + "featureImpact/",
        status=200,
        content_type="application/json",
        body=json.dumps(feature_impact_server_data),
    )

    # act
    image = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    feature_impacts = image.get_feature_impact()

    # assert
    assert feature_impacts == feature_impact_server_data["featureImpacts"]

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "featureImpact/"


@responses.activate
def test_get_feature_impact_with_metadata(
    mocked_version,
    make_versions_url,
    feature_impact_server_data,
    feature_impact_server_data_filtered,
):
    # arrange
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    mock_get_response(url, mocked_version)
    responses.add(
        responses.GET,
        url + "featureImpact/",
        status=200,
        content_type="application/json",
        body=json.dumps(feature_impact_server_data),
    )

    # act
    image = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    feature_impacts = image.get_feature_impact(with_metadata=True)

    # assert
    assert feature_impacts == feature_impact_server_data_filtered

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "featureImpact/"


@responses.activate
def test_calculate_feature_impact(unittest_endpoint, mocked_version, make_versions_url):
    # arrange
    status_url = "{}/status_url".format(unittest_endpoint)
    url = make_versions_url(mocked_version["customModelId"], mocked_version["id"])
    impact_url = url + "featureImpact/"
    mock_get_response(url, mocked_version)

    responses.add(
        responses.POST,
        impact_url,
        status=200,
        content_type="application/json",
        headers={"Location": status_url},
        # this ID is ignored, we only use headers and we dont' have a real ID from the backend.
        body=json.dumps({"status_id": "5cf4d3f5f930e26daac18a1a"}),
    )

    responses.add(
        responses.GET, status_url, headers={"Location": url}, status=303,
    )

    # act
    image = CustomModelVersion.get(mocked_version["customModelId"], mocked_version["id"])
    image.calculate_feature_impact()

    # assert

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "POST"
    assert responses.calls[1].request.url == impact_url
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == status_url
